"""archetypal DomesticHotWaterSetting."""

import collections
from statistics import mean

import numpy as np
from eppy import modeleditor
from sigfig import round
from validator_collection import validators

from archetypal import settings
from archetypal.template.schedule import UmiSchedule
from archetypal.template.umi_base import UmiBase
from archetypal.utils import log, reduce, timeit


class DomesticHotWaterSetting(UmiBase):
    """Domestic Hot Water settings.

    .. image:: ../images/template/zoneinfo-dhw.png
    """

    _CREATED_OBJECTS = []

    __slots__ = (
        "_flow_rate_per_floor_area",
        "_is_on",
        "_water_schedule",
        "_water_supply_temperature",
        "_water_temperature_inlet",
        "_area",
    )

    def __init__(
        self,
        Name,
        WaterSchedule=None,
        IsOn=True,
        FlowRatePerFloorArea=0.03,
        WaterSupplyTemperature=65,
        WaterTemperatureInlet=10,
        area=1,
        **kwargs,
    ):
        """Initialize object with parameters.

        Args:
            area (float): The area the zone associated to this object.
            IsOn (bool): If True, dhw is on.
            WaterSchedule (UmiSchedule): Schedule that modulates the
                FlowRatePerFloorArea.
            FlowRatePerFloorArea (float): The flow rate per flow area [m³/(hr·m²)].
            WaterSupplyTemperature (float): The water supply temperature [degC].
            WaterTemperatureInlet (float): The water temperature intel from the water
                mains [degC].
            **kwargs: keywords passed to parent constructors.
        """
        super(DomesticHotWaterSetting, self).__init__(Name, **kwargs)
        self.FlowRatePerFloorArea = FlowRatePerFloorArea
        self.IsOn = IsOn
        self.WaterSupplyTemperature = WaterSupplyTemperature
        self.WaterTemperatureInlet = WaterTemperatureInlet
        self.WaterSchedule = WaterSchedule
        self.area = area

        # Only at the end append self to _CREATED_OBJECTS
        self._CREATED_OBJECTS.append(self)

    @property
    def FlowRatePerFloorArea(self):
        """Get or set the flow rate per flow area [m³/(hr·m²)]."""
        return self._flow_rate_per_floor_area

    @FlowRatePerFloorArea.setter
    def FlowRatePerFloorArea(self, value):
        self._flow_rate_per_floor_area = validators.float(value, minimum=0)

    @property
    def IsOn(self):
        """Get or set the availability of the domestic hot water [bool]."""
        return self._is_on

    @IsOn.setter
    def IsOn(self, value):
        assert isinstance(value, bool), (
            f"Input error with value {value}. IsOn must "
            f"be a boolean, not a {type(value)}"
        )
        self._is_on = value

    @property
    def WaterSchedule(self):
        """Get or set the schedule which modulates the FlowRatePerFloorArea."""
        return self._water_schedule

    @WaterSchedule.setter
    def WaterSchedule(self, value):
        if value is not None:
            assert isinstance(value, UmiSchedule), (
                f"Input error with value {value}. WaterSchedule must "
                f"be an UmiSchedule, not a {type(value)}"
            )
        self._water_schedule = value

    @property
    def WaterSupplyTemperature(self):
        """Get or set the water supply temperature [degC]."""
        return self._water_supply_temperature

    @WaterSupplyTemperature.setter
    def WaterSupplyTemperature(self, value):
        self._water_supply_temperature = validators.float(value)

    @property
    def WaterTemperatureInlet(self):
        """Get or set the water temperature intel from the water mains [degC]."""
        return self._water_temperature_inlet

    @WaterTemperatureInlet.setter
    def WaterTemperatureInlet(self, value):
        self._water_temperature_inlet = validators.float(value)

    @property
    def area(self):
        """Get or set the area of the zone associated to this object [m²]."""
        return self._area

    @area.setter
    def area(self, value):
        self._area = validators.float(value, minimum=0)

    @classmethod
    def from_dict(cls, data, schedules, **kwargs):
        """Create a DomesticHotWaterSetting from a dictionary.

        Args:
            data (dict): The python dictionary.
            schedules (dict): A dictionary of UmiSchedules with their id as keys.
            **kwargs: keywords passed the MaterialBase constructor.
        """
        _id = data.pop("$id")
        wat_sch = data.pop("WaterSchedule", None)
        schedule = schedules[wat_sch["$ref"]]
        return cls(id=_id, WaterSchedule=schedule, **data, **kwargs)

    def to_dict(self):
        """Return DomesticHotWaterSetting dictionary representation."""
        self.validate()  # Validate object before trying to get json format

        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["FlowRatePerFloorArea"] = round(self.FlowRatePerFloorArea, sigfigs=4)
        data_dict["IsOn"] = self.IsOn
        data_dict["WaterSchedule"] = self.WaterSchedule.to_ref()
        data_dict["WaterSupplyTemperature"] = round(
            self.WaterSupplyTemperature, sigfigs=4
        )
        data_dict["WaterTemperatureInlet"] = round(
            self.WaterTemperatureInlet, sigfigs=4
        )
        data_dict["Category"] = self.Category
        data_dict["Comments"] = validators.string(self.Comments, allow_empty=True)
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    @classmethod
    @timeit
    def from_zone(cls, zone_epbunch, **kwargs):
        """Create object from a zone EpBunch.

        WaterUse:Equipment objects referring to this zone will be parsed.

        Args:
            zone (EpBunch): The zone object.
        """
        # If Zone is not part of Conditioned Area, it should not have a DHW object.
        if zone_epbunch.Part_of_Total_Floor_Area.lower() == "no":
            return None

        # First, find the WaterUse:Equipment assigned to this zone
        dhw_objs = zone_epbunch.getreferingobjs(
            iddgroups=["Water Systems"], fields=["Zone_Name"]
        )
        if not dhw_objs:
            # Sometimes, some of the WaterUse:Equipment objects are not assigned to
            # any zone. Therefore, to account for their water usage, we can try to
            # assign dangling WaterUse:Equipments by looking for the zone name in the
            # object name.
            dhw_objs.extend(
                [
                    dhw
                    for dhw in zone_epbunch.theidf.idfobjects["WATERUSE:EQUIPMENT"]
                    if zone_epbunch.Name.lower() in dhw.Name.lower()
                ]
            )

        if dhw_objs:
            # This zone has more than one WaterUse:Equipment object
            zone_area = modeleditor.zonearea(zone_epbunch.theidf, zone_epbunch.Name)
            total_flow_rate = cls._do_flow_rate(dhw_objs, zone_area)
            water_schedule = cls._do_water_schedule(dhw_objs)
            inlet_temp = cls._do_inlet_temp(dhw_objs)
            supply_temp = cls._do_hot_temp(dhw_objs)

            name = zone_epbunch.Name + "_DHW"
            z_dhw = cls(
                Name=name,
                FlowRatePerFloorArea=total_flow_rate,
                IsOn=bool(total_flow_rate > 0),
                WaterSchedule=water_schedule,
                WaterSupplyTemperature=supply_temp,
                WaterTemperatureInlet=inlet_temp,
                Category=zone_epbunch.theidf.name,
                area=zone_area,
                **kwargs,
            )
            return z_dhw
        else:
            log(f"No 'Water Systems' found in zone '{zone_epbunch.Name}'")
            return None

    @classmethod
    def _do_hot_temp(cls, dhw_objs):
        """Resolve hot water temperature.

        Args:
            dhw_objs:
        """
        hot_schds = []
        for obj in dhw_objs:
            # Reference to the schedule object specifying the target water
            # temperature [C]. If blank, the target temperature defaults to
            # the hot water supply temperature.
            schedule_name = (
                obj.Target_Temperature_Schedule_Name
                if obj.Target_Temperature_Schedule_Name != ""
                else obj.Hot_Water_Supply_Temperature_Schedule_Name
            )
            epbunch = obj.theidf.schedules_dict[schedule_name.upper()]
            hot_schd = UmiSchedule.from_epbunch(epbunch)
            hot_schds.append(hot_schd)

        return np.array([sched.all_values.mean() for sched in hot_schds]).mean()

    @classmethod
    def _do_inlet_temp(cls, dhw_objs):
        """Calculate inlet water temperature."""
        WaterTemperatureInlet = []
        for obj in dhw_objs:
            if obj.Cold_Water_Supply_Temperature_Schedule_Name != "":
                # If a cold water supply schedule is provided, create the
                # schedule
                epbunch = obj.theidf.schedules_dict[
                    obj.Cold_Water_Supply_Temperature_Schedule_Name.upper()
                ]
                cold_schd_names = UmiSchedule.from_epbunch(epbunch)
                WaterTemperatureInlet.append(cold_schd_names.mean)
            else:
                # If blank, water temperatures are calculated by the
                # Site:WaterMainsTemperature object.
                water_mains_temps = obj.theidf.idfobjects[
                    "Site:WaterMainsTemperature".upper()
                ]
                if water_mains_temps:
                    # If a "Site:WaterMainsTemperature" object exists,
                    # do water depending on calc method:
                    water_mains_temp = water_mains_temps[0]
                    if water_mains_temp.Calculation_Method.lower() == "schedule":
                        # From Schedule method
                        mains_scd = UmiSchedule.from_epbunch(
                            obj.theidf.schedules_dict[
                                water_mains_temp.Schedule_Name.upper()
                            ]
                        )
                        WaterTemperatureInlet.append(mains_scd.mean())
                    elif water_mains_temp.Calculation_Method.lower() == "correlation":
                        # From Correlation method
                        mean_outair_temp = (
                            water_mains_temp.Annual_Average_Outdoor_Air_Temperature
                        )
                        max_dif = (
                            water_mains_temp.Maximum_Difference_In_Monthly_Average_Outdoor_Air_Temperatures
                        )

                        WaterTemperatureInlet.append(
                            water_main_correlation(mean_outair_temp, max_dif).mean()
                        )
                else:
                    # Else, there is no Site:WaterMainsTemperature object in
                    # the input file, a default constant value of 10 C is
                    # assumed.
                    WaterTemperatureInlet.append(float(10))
        return mean(WaterTemperatureInlet) if WaterTemperatureInlet else 10

    @classmethod
    def _do_water_schedule(cls, dhw_objs):
        """Return the WaterSchedule for a list of WaterUse:Equipment objects.

        If more than one objects are passed, a combined schedule is returned

        Args:
            dhw_objs (list of EpBunch): List of WaterUse:Equipment objects.

        Returns:
            UmiSchedule: The WaterSchedule
        """
        water_schds = [
            UmiSchedule.from_epbunch(
                obj.theidf.schedules_dict[obj.Flow_Rate_Fraction_Schedule_Name.upper()],
                quantity=obj.Peak_Flow_Rate,
            )
            for obj in dhw_objs
        ]

        return reduce(
            UmiSchedule.combine,
            water_schds,
            weights=None,
            quantity=True,
        )

    @classmethod
    def _do_flow_rate(cls, dhw_objs, area):
        """Calculate total flow rate from list of WaterUse:Equipment objects.

        The zone's net_conditioned_building_area property is used to normalize the
        flow rate.

        Args:
            dhw_objs (Idf_MSequence):
        """
        total_flow_rate = 0
        for obj in dhw_objs:
            total_flow_rate += obj.Peak_Flow_Rate  # m3/s
        total_flow_rate /= area  # m3/s/m2
        total_flow_rate *= 3600.0  # m3/h/m2
        return total_flow_rate

    def combine(self, other, **kwargs):
        """Combine two DomesticHotWaterSetting objects together.

        Notes:
            When combining 2 DomesticHotWater Settings objects, the WaterSchedule
            must be averaged via the final quantity which is the peak floor rate
            [m3/hr/m2] * area [m2].

        .. code-block:: python

            (
                np.average(
                    [zone_1.WaterSchedule.all_values, zone_2.WaterSchedule.all_values],
                    axis=0,
                    weights=[
                        zone_1.FlowRatePerFloorArea * zone_1.area,
                        zone_2.FlowRatePerFloorArea * zone_2.area,
                    ],
                )
                * (combined.FlowRatePerFloorArea * 100)
            ).sum()

        Args:
            other (DomesticHotWaterSetting): The other object.
            **kwargs: keywords passed to the constructor.

        Returns:
            (DomesticHotWaterSetting): a new combined object.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        meta = self._get_predecessors_meta(other)
        new_obj = DomesticHotWaterSetting(
            WaterSchedule=UmiSchedule.combine(
                self.WaterSchedule,
                other.WaterSchedule,
                weights=[
                    self.FlowRatePerFloorArea * self.area,
                    other.FlowRatePerFloorArea * other.area,
                ],
            ),
            IsOn=any((self.IsOn, other.IsOn)),
            FlowRatePerFloorArea=self.float_mean(
                other, "FlowRatePerFloorArea", [self.area, other.area]
            ),
            WaterSupplyTemperature=self.float_mean(
                other, "WaterSupplyTemperature", [self.area, other.area]
            ),
            WaterTemperatureInlet=self.float_mean(
                other, "WaterTemperatureInlet", [self.area, other.area]
            ),
            area=self.area + other.area,
            **meta,
        )
        new_obj.predecessors.update(self.predecessors + other.predecessors)
        return new_obj

    def validate(self):
        """Validate object and fill in missing values."""
        return self

    @classmethod
    def whole_building(cls, idf):
        """Create one DomesticHotWaterSetting for whole building model.

        Args:
            idf (IDF): The idf model.

        Returns:
            DomesticHotWaterSetting: The DomesticHotWaterSetting object.
        """
        # Unconditioned area could be zero, therefore taking max of both
        area = max(idf.net_conditioned_building_area, idf.unconditioned_building_area)

        z_dhw_list = []
        dhw_objs = idf.idfobjects["WaterUse:Equipment".upper()]
        if not dhw_objs:
            # defaults with 0 flow rate.
            total_flow_rate = 0
            water_schedule = UmiSchedule.constant_schedule()
            supply_temp = 60
            inlet_temp = 10

            name = idf.name + "_DHW"
            z_dhw = DomesticHotWaterSetting(
                WaterSchedule=water_schedule,
                IsOn=bool(total_flow_rate > 0),
                FlowRatePerFloorArea=total_flow_rate,
                WaterSupplyTemperature=supply_temp,
                WaterTemperatureInlet=inlet_temp,
                area=area,
                Name=name,
                Category=idf.name,
            )
            z_dhw_list.append(z_dhw)
        else:
            total_flow_rate = DomesticHotWaterSetting._do_flow_rate(dhw_objs, area)
            water_schedule = DomesticHotWaterSetting._do_water_schedule(dhw_objs)
            water_schedule.quantity = total_flow_rate
            inlet_temp = DomesticHotWaterSetting._do_inlet_temp(dhw_objs)
            supply_temp = DomesticHotWaterSetting._do_hot_temp(dhw_objs)
            z_dhw = DomesticHotWaterSetting(
                WaterSchedule=water_schedule,
                IsOn=bool(total_flow_rate > 0),
                FlowRatePerFloorArea=total_flow_rate,
                WaterSupplyTemperature=supply_temp,
                WaterTemperatureInlet=inlet_temp,
                area=area,
                Name="Whole Building WaterUse:Equipment",
                Category=idf.name,
            )
            z_dhw_list.append(z_dhw)

        return reduce(DomesticHotWaterSetting.combine, z_dhw_list)

    def mapping(self, validate=False):
        """Get a dict based on the object properties, useful for dict repr.

        Args:
            validate (bool): If True, try to validate object before returning the
                mapping.
        """
        if validate:
            self.validate()

        return dict(
            FlowRatePerFloorArea=self.FlowRatePerFloorArea,
            IsOn=self.IsOn,
            WaterSchedule=self.WaterSchedule,
            WaterSupplyTemperature=self.WaterSupplyTemperature,
            WaterTemperatureInlet=self.WaterTemperatureInlet,
            Category=self.Category,
            Comments=self.Comments,
            DataSource=self.DataSource,
            Name=self.Name,
        )

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __add__(self, other):
        """Overload + to implement self.combine.

        Args:
            other (DomesticHotWaterSetting):
        """
        return self.combine(
            other,
        )

    def __hash__(self):
        """Return the hash value of self."""
        return hash(self.id)

    def __key__(self):
        """Get a tuple of attributes. Useful for hashing and comparing."""
        return (
            self.IsOn,
            self.FlowRatePerFloorArea,
            self.WaterSupplyTemperature,
            self.WaterTemperatureInlet,
            self.WaterSchedule,
        )

    def __eq__(self, other):
        """Assert self is equivalent to other."""
        if not isinstance(other, DomesticHotWaterSetting):
            return NotImplemented
        else:
            return self.__key__() == other.__key__()

    def __str__(self):
        """Return string representation."""
        return (
            f"{str(self.id)}: {str(self.Name)} "
            f"PeakFlow {self.FlowRatePerFloorArea:.5f} m3/hr/m2"
        )

    def __copy__(self):
        """Create a copy of self."""
        return self.__class__(**self.mapping(validate=False))

    @property
    def children(self):
        return (self.WaterSchedule,)

    def to_epbunch(self, idf, zone_name, zone_area):
        """Convert self to the EpBunches given an idf model, a zone name.

        Notes:
            'Target_Temperature_Schedule_Name' Reference to the schedule object specifying the
            target water temperature [C]. Hot and cold water are mixed at the tap to attain
            the target temperature. If insufficient hot water is available to reach the target
            temperature, the result is cooler water at the tap. If blank, the target
            temperature defaults to the hot water supply temperature.

        Args:
            idf (IDF): The idf model in which the EpBunch is created.
            zone_name (str): The zone name to associate this EpBunch (str)
            zone_area (float): The occupied zone floor area (m2)

        .. code-block::

            WATERUSE:EQUIPMENT,
                DHW Perim,                !- Name
                DHW Perim,                !- EndUse Subcategory
                2.796438e-07,             !- Peak Flow Rate
                B_Off_Y_Dhw,              !- Flow Rate Fraction Schedule Name
                DHW Supply Temperature Perim,    !- Target Temperature Schedule Name
                DHW Supply Temperature Perim,    !- Hot Water Supply Temperature Schedule Name
                DHW Mains Temperature Perim,    !- Cold Water Supply Temperature Schedule Name
                ,                         !- Zone Name
                ,                         !- Sensible Fraction Schedule Name
                ;                         !- Latent Fraction Schedule Name

        Returns:
            tuple: An EpBunch object added to the idf model.
        """
        if self.IsOn:
            # Create ep_bunch for zone
            dhw_epbunch = idf.newidfobject(
                key="WATERUSE:EQUIPMENT",
                Name=f"{zone_name} DHW",
                EndUse_Subcategory="DHW",
                # TODO: check if this needs to be a string in sci not (self.FlowRatePerFloorArea)
                Peak_Flow_Rate=self.FlowRatePerFloorArea / 3600 * zone_area,  # m3/s
                Flow_Rate_Fraction_Schedule_Name=self.WaterSchedule.to_epbunch(
                    idf
                ).Name,  # self.WaterSchedule.to_epbunch(idf).Name
                Target_Temperature_Schedule_Name="",
                Hot_Water_Supply_Temperature_Schedule_Name=idf.newidfobject(
                    key="SCHEDULE:CONSTANT",
                    Name="DhwHotTemp",
                    Hourly_Value=self.WaterSupplyTemperature,
                ).Name,
                Cold_Water_Supply_Temperature_Schedule_Name=idf.newidfobject(
                    key="SCHEDULE:CONSTANT",
                    Name="DhwInletTemp",
                    Hourly_Value=self.WaterTemperatureInlet,
                ).Name,
                Zone_Name=zone_name,
                Sensible_Fraction_Schedule_Name="",
                Latent_Fraction_Schedule_Name="",
            )
        else:
            infiltration_epbunch = None
            log("No 'WATERUSE:EQUIPMENT' created since DHW IsOn == False.")
        return dhw_epbunch


def water_main_correlation(t_out_avg, max_diff):
    """Based on the correlation developed by Craig Christensen and Jay Burch.

    Returns a 365 days temperature profile.

    Info:
        https://bigladdersoftware.com/epx/docs/8-9/engineering-reference
        /water-systems.html#water-mains-temperatures

    Args:
        t_out_avg (float): average annual outdoor air temperature (°C).
        max_diff (float): maximum difference in monthly average outdoor air
            temperatures (°C).

    Returns:
        (pd.Series): water mains temperature profile.
    """
    import numpy as np
    import pandas as pd

    Q_ = settings.unit_registry.Quantity
    t_out_avg_F = Q_(t_out_avg, "degC").to("degF")
    max_diff_F = Q_(max_diff, "delta_degC").to("delta_degF")
    ratio = 0.4 + 0.01 * (t_out_avg_F.m - 44)
    lag = 35 - 1.0 * (t_out_avg_F.m - 44)
    days = np.arange(1, 365)

    def function(t_out_avg, day, max_diff):
        return (t_out_avg + 6) + ratio * (max_diff / 2) * np.sin(
            np.deg2rad(0.986 * (day - 15 - lag) - 90)
        )

    mains = [Q_(function(t_out_avg_F.m, day, max_diff_F.m), "degF") for day in days]
    series = pd.Series([temp.to("degC").m for temp in mains])
    return series



"""archetypal Schedule module."""

import functools
import io
import logging as lg
from datetime import datetime, timedelta
from itertools import groupby
from typing import FrozenSet, List, Union

import numpy as np
import pandas as pd
from energy_pandas import EnergySeries
from eppy.bunch_subclass import BadEPFieldError
from typing_extensions import Literal
from validator_collection import checkers, validators

from archetypal.utils import log


class ScheduleTypeLimits:
    """ScheduleTypeLimits class."""

    __slots__ = ("_name", "_lower_limit", "_upper_limit", "_numeric_type", "_unit_type")

    _NUMERIC_TYPES = ("continuous", "discrete")
    _UNIT_TYPES = (
        "Dimensionless",
        "Temperature",
        "DeltaTemperature",
        "PrecipitationRate",
        "Angle",
        "ConvectionCoefficient",
        "ActivityLevel",
        "Velocity",
        "Capacity",
        "Power",
        "Availability",
        "Percent",
        "Control",
        "Mode",
    )

    def __init__(
        self,
        Name,
        LowerLimit,
        UpperLimit,
        NumericType="Continuous",
        UnitType="Dimensionless",
    ):
        """Initialize object."""
        self.Name = Name
        self.LowerLimit = LowerLimit
        self.UpperLimit = UpperLimit
        self.NumericType = NumericType
        self.UnitType = UnitType

    @property
    def Name(self):
        """Get or set the name of the ScheduleTypeLimits."""
        return self._name

    @Name.setter
    def Name(self, value):
        self._name = validators.string(value)

    @property
    def LowerLimit(self):
        """Get or set the LowerLimit."""
        return self._lower_limit

    @LowerLimit.setter
    def LowerLimit(self, value):
        self._lower_limit = validators.float(value, allow_empty=True)

    @property
    def UpperLimit(self):
        """Get or set the UpperLimit."""
        return self._upper_limit

    @UpperLimit.setter
    def UpperLimit(self, value):
        self._upper_limit = validators.float(value, allow_empty=True)

    @property
    def NumericType(self):
        """Get or set numeric type. Can be null."""
        return self._numeric_type

    @NumericType.setter
    def NumericType(self, value):
        validators.string(value, allow_empty=True)
        if value is not None:
            assert value.lower() in self._NUMERIC_TYPES, (
                f"Input error for value '{value}'. NumericType must "
                f"be one of '{self._NUMERIC_TYPES}'"
            )
        self._numeric_type = value

    @property
    def UnitType(self):
        """Get or set the unit type. Can be null."""
        return self._unit_type

    @UnitType.setter
    def UnitType(self, value):
        value = validators.string(value)
        assert value.lower() in map(str.lower, self._UNIT_TYPES), (
            f"Input error for value '{value}'. UnitType must "
            f"be one of '{self._UNIT_TYPES}'"
        )
        self._unit_type = value

    @classmethod
    def from_dict(cls, data):
        """Create a ScheduleTypeLimit from a dictionary.

        Args:
            data: ScheduleTypeLimit dictionary following the format below.

        .. code-block:: python

            {
            "Name": 'Fractional',
            "LowerLimit": 0,
            "UpperLimit": 1,
            "NumericType": None,
            "UnitType": "Dimensionless"
            }
        """
        return cls(**data)

    @classmethod
    def from_epbunch(cls, epbunch):
        """Create a ScheduleTypeLimits from an epbunch.

        Args:
            epbunch (EpBunch): The epbunch of key "SCHEDULETYPELIMITS".
        """
        assert (
            epbunch.key.upper() == "SCHEDULETYPELIMITS"
        ), f"Expected 'SCHEDULETYPELIMITS' epbunch. Got {epbunch.key}."
        name = epbunch.Name
        lower_limit = epbunch.Lower_Limit_Value
        upper_limit = epbunch.Upper_Limit_Value
        numeric_type = epbunch.Numeric_Type
        unit_type = epbunch.Unit_Type
        return cls(
            Name=name,
            LowerLimit=lower_limit if checkers.is_numeric(lower_limit) else None,
            UpperLimit=upper_limit if checkers.is_numeric(upper_limit) else None,
            NumericType=numeric_type
            if checkers.is_string(numeric_type, minimum_length=1)
            else "Continuous",
            UnitType=unit_type
            if checkers.is_string(unit_type, minimum_length=1)
            else "Dimensionless",
        )

    def to_dict(self):
        """Return ScheduleTypeLimits dictionary representation."""
        return {
            "Name": self.Name,
            "LowerLimit": self.LowerLimit,
            "UpperLimit": self.UpperLimit,
            "NumericType": self.NumericType,
            "UnitType": self.UnitType,
        }

    def to_epbunch(self, idf):
        """Convert self to an epbunch given an idf model.

        Notes:
            The object is added to the idf model.

        Args:
            idf (IDF): An IDF model.

        .. code-block:: python

            SCHEDULETYPELIMITS,
                ,                         !- Name
                ,                         !- Lower Limit Value
                ,                         !- Upper Limit Value
                ,                         !- Numeric Type
                Dimensionless;            !- Unit Type

        Returns:
            EpBunch: The EpBunch object added to the idf model.
        """
        return idf.newidfobject(
            key="SCHEDULETYPELIMITS",
            Name=self.Name,
            Lower_Limit_Value=self.LowerLimit,
            Upper_Limit_Value=self.UpperLimit,
            Numeric_Type=self.NumericType,
            Unit_Type=self.UnitType,
        )

    def duplicate(self):
        """Get copy of self."""
        return self.__copy__()

    def __copy__(self):
        """Get copy of self."""
        return self.__class__(
            self.Name, self.LowerLimit, self.UpperLimit, self.NumericType, self.UnitType
        )

    def __repr__(self):
        """Return the string representation of self."""
        return (
            self.Name
            + f" {self.LowerLimit} < values < {self.UpperLimit}"
            + f"Units: {self.UnitType}"
        )

    def __keys__(self):
        """Get keys of self. Useful for hashing."""
        return (
            self.Name,
            self.LowerLimit,
            self.UpperLimit,
            self.NumericType,
            self.UnitType,
        )

    def __eq__(self, other):
        """Assert self is equal to other."""
        if not isinstance(other, ScheduleTypeLimits):
            return NotImplemented
        else:
            return self.__keys__() == other.__keys__()


class _ScheduleParser:
    """Class used to parse schedules in IDF files."""

    @staticmethod
    def get_interval_day_ep_schedule_values(epbunch) -> np.ndarray:
        """Get values for Schedule:Day:Interval.

        Args:
            epbunch (EpBunch): The schedule EpBunch object.
        """
        number_of_day_sch = int((len(epbunch.fieldvalues) - 3) / 2)

        hourly_values = np.arange(24, dtype=float)
        start_hour = 0
        for i in range(number_of_day_sch):
            value = float(epbunch["Value_Until_Time_{}".format(i + 1)])
            until_time = [
                int(s.strip())
                for s in epbunch["Time_{}".format(i + 1)].split(":")
                if s.strip().isdigit()
            ]
            end_hour = int(until_time[0] + until_time[1] / 60)
            for hour in range(start_hour, end_hour):
                hourly_values[hour] = value

            start_hour = end_hour
        _, _, numeric_type, _ = _ScheduleParser.get_schedule_type_limits_data(epbunch)
        if numeric_type.strip().lower() == "discrete":
            hourly_values = hourly_values.astype(int)

        return hourly_values

    @staticmethod
    def get_hourly_day_ep_schedule_values(epbunch):
        """Get values for Schedule:Day:Hourly.

        Args:
            epbunch (EpBunch): The schedule EpBunch object.
        """
        return np.array(epbunch.fieldvalues[3:])

    @staticmethod
    def get_compact_weekly_ep_schedule_values(
        epbunch, start_date, index=None, strict=False
    ) -> np.ndarray:
        """Get values for schedule:week:compact.

        Args:
            strict:
            epbunch (EpBunch): the name of the schedule
            start_date:
            index:
        """
        if index is None:
            idx = pd.date_range(start=start_date, periods=168, freq="1H")
            slicer_ = pd.Series([False] * (len(idx)), index=idx)
        else:
            slicer_ = pd.Series([False] * (len(index)), index=index)

        weekly_schedules = pd.Series([0] * len(slicer_), index=slicer_.index)
        # update last day of schedule

        num_of_daily_schedules = int(len(epbunch.fieldvalues[2:]) / 2)

        for i in range(num_of_daily_schedules):
            day_type = epbunch["DayType_List_{}".format(i + 1)].lower()
            # This field can optionally contain the prefix “For”
            how = _ScheduleParser._field_set(
                epbunch, day_type.strip("for: "), start_date, slicer_, strict
            )
            if not weekly_schedules.loc[how].empty:
                # Loop through days and replace with day:schedule values
                days = []
                for name, day in weekly_schedules.loc[how].groupby(
                    pd.Grouper(freq="D")
                ):
                    if not day.empty:
                        ref = epbunch.get_referenced_object(
                            "ScheduleDay_Name_{}".format(i + 1)
                        )
                        day.loc[:] = _ScheduleParser.get_schedule_values(
                            sched_epbunch=ref, start_date=start_date, strict=strict
                        )
                        days.append(day)
                new = pd.concat(days)
                slicer_.update(pd.Series([True] * len(new.index), index=new.index))
                slicer_ = slicer_.apply(lambda x: x is True)
                weekly_schedules.update(new)
            else:
                return weekly_schedules.values

        return weekly_schedules.values

    @staticmethod
    def get_daily_weekly_ep_schedule_values(epbunch, start_date, strict) -> np.ndarray:
        """Get values for schedule:week:daily.

        Args:
            strict:
            epbunch (EpBunch): The schedule EpBunch object.
        """
        # 7 list for 7 days of the week
        hourly_values = []
        for day in [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]:
            ref = epbunch.get_referenced_object("{}_ScheduleDay_Name".format(day))
            h = _ScheduleParser.get_schedule_values(
                sched_epbunch=ref, start_date=start_date, strict=strict
            )
            hourly_values.append(h)
        hourly_values = np.array(hourly_values)
        # shift days earlier by self.startDayOfTheWeek
        hourly_values = np.roll(hourly_values, -start_date.weekday(), axis=0)

        return hourly_values.ravel()

    @staticmethod
    def get_list_day_ep_schedule_values(epbunch, start_date) -> np.ndarray:
        """Get values for schedule:day:list.

        Args:
            start_date:
            epbunch (EpBunch): The schedule epbunch object.
        """
        freq = int(epbunch["Minutes_per_Item"])  # Frequency of the values
        num_values = epbunch.fieldvalues[5:]  # List of values
        method = epbunch["Interpolate_to_Timestep"]  # How to resample

        # fill a list of available values and pad with zeros (this is safer
        # but should not occur)
        all_values = np.arange(int(24 * 60 / freq))
        for i in all_values:
            try:
                all_values[i] = num_values[i]
            except Exception:
                all_values[i] = 0
        # create a fake index to help us with the resampling
        index = pd.date_range(
            start=start_date, periods=(24 * 60) / freq, freq="{}T".format(freq)
        )
        series = pd.Series(all_values, index=index)

        # resample series to hourly values and apply resampler function
        series = series.resample("1H").apply(_how(method))

        return series.values

    @staticmethod
    def get_constant_ep_schedule_values(epbunch) -> np.ndarray:
        """Get values for schedule:constant.

        Args:
            epbunch (EpBunch): The schedule epbunch object.
        """
        hourly_values = np.arange(8760)
        value = float(epbunch["Hourly_Value"])
        for hour in hourly_values:
            hourly_values[hour] = value
        _, _, numeric_type, _ = _ScheduleParser.get_schedule_type_limits_data(epbunch)
        if numeric_type.strip().lower() == "discrete":
            hourly_values = hourly_values.astype(int)

        return hourly_values

    @staticmethod
    def get_file_ep_schedule_values(epbunch) -> np.ndarray:
        """Get values for schedule:file.

        Args:
            epbunch (EpBunch): The schedule epbunch object.
        """
        filename = epbunch["File_Name"]
        column = epbunch["Column_Number"]
        rows = epbunch["Rows_to_Skip_at_Top"]
        # hours = epbunch["Number_of_Hours_of_Data"]
        sep = epbunch["Column_Separator"]
        # interp = epbunch["Interpolate_to_Timestep"]

        file = epbunch.theidf.simulation_dir.files(filename)[0]

        delimeter = _separator(sep)
        skip_rows = int(rows) - 1  # We want to keep the column
        col = [int(column) - 1]  # zero-based
        epbunch = pd.read_csv(
            file, delimiter=delimeter, skiprows=skip_rows, usecols=col
        )

        return epbunch.iloc[:, 0].values

    @staticmethod
    def get_compact_ep_schedule_values(epbunch, start_date, strict) -> np.ndarray:
        """Get values for schedule:compact.

        Args:
            strict:
            start_date:
            epbunch (EpBunch): The schedule epbunch object.
        """
        field_sets = ["through", "for", "interpolate", "until", "value"]
        fields = epbunch.fieldvalues[3:]

        index = pd.date_range(start=start_date, periods=8760, freq="H")
        zeros = np.zeros(len(index))

        slicer_ = pd.Series([False] * len(index), index=index)
        series = pd.Series(zeros, index=index)

        from_day = start_date
        ep_from_day = datetime(start_date.year, 1, 1)
        from_time = "00:00"
        how_interpolate = None
        for field in fields:
            if any([spe in field.lower() for spe in field_sets]):
                f_set, hour, minute, value = _ScheduleParser._field_interpreter(
                    field, epbunch.Name
                )

                if f_set.lower() == "through":
                    # main condition. All sub-conditions must obey a
                    # `Through` condition

                    # First, initialize the slice (all False for now)
                    through_conditions = _ScheduleParser._invalidate_condition(series)

                    # reset from_time
                    from_time = "00:00"

                    # Prepare ep_to_day variable
                    ep_to_day = _ScheduleParser._date_field_interpretation(
                        value, start_date
                    ) + timedelta(days=1)

                    # Calculate Timedelta in days
                    days = (ep_to_day - ep_from_day).days
                    # Add timedelta to startDate
                    to_day = from_day + timedelta(days=days) + timedelta(hours=-1)

                    # slice the conditions with the range and apply True
                    through_conditions.loc[from_day:to_day] = True

                    from_day = to_day + timedelta(hours=1)
                    ep_from_day = ep_to_day
                elif f_set.lower() == "for":
                    # slice specific days
                    # reset from_time
                    from_time = "00:00"

                    for_condition = _ScheduleParser._invalidate_condition(series)
                    fors = value.split()
                    if len(fors) > 1:
                        # if multiple `For`. eg.: For: Weekends Holidays,
                        # Combine all conditions
                        for value in fors:
                            if value.lower() == "allotherdays":
                                # Apply condition to slice
                                how = _ScheduleParser._field_set(
                                    epbunch, value, start_date, slicer_, strict
                                )
                                # Reset for condition
                                for_condition = how
                            else:
                                how = _ScheduleParser._field_set(
                                    epbunch, value, start_date, slicer_, strict
                                )
                                if how is not None:
                                    for_condition.loc[how] = True
                    elif value.lower() == "allotherdays":
                        # Apply condition to slice
                        how = _ScheduleParser._field_set(
                            epbunch, value, start_date, slicer_, strict
                        )
                        # Reset for condition
                        for_condition = how
                    else:
                        # Apply condition to slice
                        how = _ScheduleParser._field_set(
                            epbunch, value, start_date, slicer_, strict
                        )
                        for_condition.loc[how] = True

                    # Combine the for_condition with all_conditions
                    all_conditions = through_conditions & for_condition

                    # update in memory slice
                    # self.sliced_day_.loc[all_conditions] = True
                elif "interpolate" in f_set.lower():
                    # we need to upsample to series to 8760 * 60 values
                    new_idx = pd.date_range(
                        start=start_date, periods=525600, closed="left", freq="T"
                    )
                    series = series.resample("T").pad()
                    series = series.reindex(new_idx)
                    series.fillna(method="pad", inplace=True)
                    through_conditions = through_conditions.resample("T").pad()
                    through_conditions = through_conditions.reindex(new_idx)
                    through_conditions.fillna(method="pad", inplace=True)
                    for_condition = for_condition.resample("T").pad()
                    for_condition = for_condition.reindex(new_idx)
                    for_condition.fillna(method="pad", inplace=True)
                    how_interpolate = value.lower()
                elif f_set.lower() == "until":
                    until_condition = _ScheduleParser._invalidate_condition(series)
                    if series.index.freq.name == "T":
                        # until_time = str(int(hour) - 1) + ':' + minute
                        until_time = timedelta(
                            hours=int(hour), minutes=int(minute)
                        ) - timedelta(minutes=1)

                    else:
                        until_time = str(int(hour) - 1) + ":" + minute
                    until_condition.loc[
                        until_condition.between_time(from_time, str(until_time)).index
                    ] = True
                    all_conditions = (
                        for_condition & through_conditions & until_condition
                    )

                    from_time = str(int(hour)) + ":" + minute
                elif f_set.lower() == "value":
                    # If the therm `Value: ` field is used, we will catch it
                    # here.
                    # update in memory slice
                    slicer_.loc[all_conditions] = True
                    series[all_conditions] = value
                else:
                    # Do something here before looping to the next Field
                    pass
            else:
                # If the term `Value: ` is not used; the variable is simply
                # passed in the Field
                value = float(field)
                series[all_conditions] = value

                # update in memory slice
                slicer_.loc[all_conditions] = True
        if how_interpolate:
            return series.resample("H").mean().values
        else:
            return series.values

    @classmethod
    def get_yearly_ep_schedule_values(cls, epbunch, start_date, strict) -> np.ndarray:
        """Get values for schedule:year.

        Args:
            strict:
            start_date (datetime):
            epbunch (EpBunch): the schedule epbunch.
        """
        # first week
        year = start_date.year
        idx = pd.date_range(start=start_date, periods=8760, freq="1H")
        hourly_values = pd.Series([0] * 8760, index=idx)

        # generate weekly schedules
        num_of_weekly_schedules = int(len(epbunch.fieldvalues[3:]) / 5)

        for i in range(num_of_weekly_schedules):
            ref = epbunch.get_referenced_object("ScheduleWeek_Name_{}".format(i + 1))

            start_month = getattr(epbunch, "Start_Month_{}".format(i + 1))
            end_month = getattr(epbunch, "End_Month_{}".format(i + 1))
            start_day = getattr(epbunch, "Start_Day_{}".format(i + 1))
            end_day = getattr(epbunch, "End_Day_{}".format(i + 1))

            start = datetime.strptime(
                "{}/{}/{}".format(year, start_month, start_day), "%Y/%m/%d"
            )
            end = datetime.strptime(
                "{}/{}/{}".format(year, end_month, end_day), "%Y/%m/%d"
            )
            days = (end - start).days + 1

            end_date = start_date + timedelta(days=days) + timedelta(hours=23)
            how = pd.IndexSlice[start_date:end_date]

            weeks = []
            for name, week in hourly_values.loc[how].groupby(pd.Grouper(freq="168H")):
                if not week.empty:
                    try:
                        week.loc[:] = cls.get_schedule_values(
                            sched_epbunch=ref,
                            start_date=week.index[0],
                            index=week.index,
                            strict=strict,
                        )
                    except ValueError:
                        week.loc[:] = cls.get_schedule_values(
                            sched_epbunch=ref, start_date=week.index[0], strict=strict
                        )[0 : len(week)]
                    finally:
                        weeks.append(week)
            new = pd.concat(weeks)
            hourly_values.update(new)
            start_date += timedelta(days=days)

        return hourly_values.values

    @staticmethod
    def get_schedule_values(
        sched_epbunch, start_date, index=None, strict=False
    ) -> list:
        """Get schedule values for epbunch.

        Args:
            strict:
            sched_epbunch (EpBunch): the schedule epbunch object
            start_date:
            index:
        """
        cls = _ScheduleParser
        sch_type = sched_epbunch.key.upper()

        if sch_type.upper() == "schedule:year".upper():
            hourly_values = cls.get_yearly_ep_schedule_values(
                sched_epbunch, start_date, strict
            )
        elif sch_type.upper() == "schedule:day:interval".upper():
            hourly_values = cls.get_interval_day_ep_schedule_values(sched_epbunch)
        elif sch_type.upper() == "schedule:day:hourly".upper():
            hourly_values = cls.get_hourly_day_ep_schedule_values(sched_epbunch)
        elif sch_type.upper() == "schedule:day:list".upper():
            hourly_values = cls.get_list_day_ep_schedule_values(
                sched_epbunch, start_date
            )
        elif sch_type.upper() == "schedule:week:compact".upper():
            hourly_values = cls.get_compact_weekly_ep_schedule_values(
                sched_epbunch, start_date, index, strict
            )
        elif sch_type.upper() == "schedule:week:daily".upper():
            hourly_values = cls.get_daily_weekly_ep_schedule_values(
                sched_epbunch, start_date, strict
            )
        elif sch_type.upper() == "schedule:constant".upper():
            hourly_values = cls.get_constant_ep_schedule_values(sched_epbunch)
        elif sch_type.upper() == "schedule:compact".upper():
            hourly_values = cls.get_compact_ep_schedule_values(
                sched_epbunch, start_date, strict
            )
        elif sch_type.upper() == "schedule:file".upper():
            hourly_values = cls.get_file_ep_schedule_values(sched_epbunch)
        else:
            log(
                "Archetypal does not currently support schedules of type "
                '"{}"'.format(sch_type),
                lg.WARNING,
            )
            hourly_values = []

        return list(hourly_values)

    @staticmethod
    def _field_interpreter(field, name):
        """Deal with a Field-Set (Through, For, Interpolate, # Until, Value).

        Args:
            name:
            field:

        Returns:
            string: the parsed string
        """
        values_sets = [
            "weekdays",
            "weekends",
            "alldays",
            "allotherdays",
            "sunday",
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "summerdesignday",
            "winterdesignday",
            "holiday",
        ]
        keywords = None

        if "through" in field.lower():
            # deal with through
            if ":" in field.lower():
                # parse colon
                f_set, statement = field.split(":")
                hour = None
                minute = None
                value = statement.strip()
            else:
                msg = (
                    'The schedule "{sch}" contains a Field '
                    'that is not understood: "{field}"'.format(sch=name, field=field)
                )
                raise NotImplementedError(msg)
        elif "for" in field.lower():
            keywords = [word for word in values_sets if word in field.lower()]
            if ":" in field.lower():
                # parse colon
                f_set, statement = field.split(":")
                value = statement.strip()
                hour = None
                minute = None
            elif keywords:
                # get epBunch of the sizing period
                statement = " ".join(keywords)
                f_set = [s for s in field.split() if "for" in s.lower()][0]
                value = statement.strip()
                hour = None
                minute = None
            else:
                # parse without a colon
                msg = (
                    'The schedule "{sch}" contains a Field '
                    'that is not understood: "{field}"'.format(sch=name, field=field)
                )
                raise NotImplementedError(msg)
        elif "interpolate" in field.lower():
            msg = (
                'The schedule "{sch}" contains sub-hourly values ('
                'Field-Set="{field}"). The average over the hour is '
                "taken".format(sch=name, field=field)
            )
            log(msg, lg.WARNING)
            f_set, value = field.split(":")
            hour = None
            minute = None
        elif "until" in field.lower():
            if ":" in field.lower():
                # parse colon
                try:
                    f_set, hour, minute = field.split(":")
                    hour = hour.strip()  # remove trailing spaces
                    minute = minute.strip()  # remove trailing spaces
                    value = None
                except Exception:
                    f_set = "until"
                    hour, minute = field.split(":")
                    hour = hour[-2:].strip()
                    minute = minute.strip()
                    value = None
            else:
                msg = (
                    'The schedule "{sch}" contains a Field '
                    'that is not understood: "{field}"'.format(sch=name, field=field)
                )
                raise NotImplementedError(msg)
        elif "value" in field.lower():
            if ":" in field.lower():
                # parse colon
                f_set, statement = field.split(":")
                value = statement.strip()
                hour = None
                minute = None
            else:
                msg = (
                    'The schedule "{sch}" contains a Field '
                    'that is not understood: "{field}"'.format(sch=name, field=field)
                )
                raise NotImplementedError(msg)
        else:
            # deal with the data value
            f_set = field
            hour = None
            minute = None
            value = field[len(field) + 1 :].strip()

        return f_set, hour, minute, value

    @staticmethod
    def _invalidate_condition(series):
        index = series.index
        periods = len(series)
        return pd.Series([False] * periods, index=index)

    @staticmethod
    def get_schedule_type_limits_data(epbunch):
        """Return schedule type limits info for epbunch."""
        try:
            schedule_limit_name = epbunch.Schedule_Type_Limits_Name
        except Exception:
            # this schedule is probably a 'Schedule:Week:Daily' which does
            # not have a Schedule_Type_Limits_Name field
            return "", "", "", ""
        else:
            (
                lower_limit,
                upper_limit,
                numeric_type,
                unit_type,
            ) = epbunch.theidf.get_schedule_type_limits_data_by_name(
                schedule_limit_name
            )

            return lower_limit, upper_limit, numeric_type, unit_type

    @staticmethod
    def _field_set(schedule_epbunch, field, start_date, slicer_=None, strict=False):
        """Return the proper slicer depending on the _field_set value.

        Available values are: Weekdays, Weekends, Holidays, Alldays,
        SummerDesignDay, WinterDesignDay, Sunday, Monday, Tuesday, Wednesday,
        Thursday, Friday, Saturday, CustomDay1, CustomDay2, AllOtherDays

        Args:
            start_date:
            strict:
            schedule_epbunch:
            field (str): The EnergyPlus field set value.
            slicer_:

        Returns:
            (indexer-like): Returns the appropriate indexer for the series.
        """
        if field.lower() == "weekdays":
            # return only days of weeks
            return lambda x: x.index.dayofweek < 5
        elif field.lower() == "weekends":
            # return only weekends
            return lambda x: x.index.dayofweek >= 5
        elif field.lower() == "alldays":
            # return all days := equivalent to .loc[:]
            return pd.IndexSlice[:]
        elif field.lower() == "allotherdays":
            # return unused days (including special days). Uses the global
            # variable `slicer_`
            import operator

            if slicer_ is not None:
                return _conjunction(
                    *[
                        _ScheduleParser.special_day(
                            schedule_epbunch, field, slicer_, strict, start_date
                        ),
                        ~slicer_,
                    ],
                    logical=operator.or_,
                )
            else:
                raise NotImplementedError
        elif field.lower() == "sunday":
            # return only sundays
            return lambda x: x.index.dayofweek == 6
        elif field.lower() == "monday":
            # return only mondays
            return lambda x: x.index.dayofweek == 0
        elif field.lower() == "tuesday":
            # return only Tuesdays
            return lambda x: x.index.dayofweek == 1
        elif field.lower() == "wednesday":
            # return only Wednesdays
            return lambda x: x.index.dayofweek == 2
        elif field.lower() == "thursday":
            # return only Thursdays
            return lambda x: x.index.dayofweek == 3
        elif field.lower() == "friday":
            # return only Fridays
            return lambda x: x.index.dayofweek == 4
        elif field.lower() == "saturday":
            # return only Saturdays
            return lambda x: x.index.dayofweek == 5
        elif field.lower() == "summerdesignday":
            # return _ScheduleParser.design_day(
            #     schedule_epbunch, field, slicer_, start_date, strict
            # )
            return None
        elif field.lower() == "winterdesignday":
            # return _ScheduleParser.design_day(
            #     schedule_epbunch, field, slicer_, start_date, strict
            # )
            return None
        elif field.lower() == "holiday" or field.lower() == "holidays":
            field = "holiday"
            return _ScheduleParser.special_day(
                schedule_epbunch, field, slicer_, strict, start_date
            )
        elif not strict:
            # If not strict, ignore missing field-sets such as CustomDay1
            return lambda x: x < 0
        else:
            raise NotImplementedError(
                f"Archetypal does not yet support The Field_set '{field}'"
            )

    @staticmethod
    def _date_field_interpretation(field, start_date):
        """Date Field Interpretation.

        Info:
            See EnergyPlus documentation for more details: 1.6.8.1.2 Field:
            Start Date (Table 1.4: Date Field Interpretation)

        Args:
            start_date:
            field (str): The EnergyPlus Field Contents

        Returns:
            (datetime): The datetime object
        """
        # < number > Weekday in Month
        formats = ["%m/%d", "%d %B", "%B %d", "%d %b", "%b %d"]
        date = None
        for format_str in formats:
            # Tru to parse using each defined formats
            try:
                date = datetime.strptime(field, format_str)
            except Exception:
                pass
            else:
                date = datetime(start_date.year, date.month, date.day)
        if date is None:
            # if the defined formats did not work, try the fancy parse
            try:
                date = _ScheduleParser._parse_fancy_string(field, start_date)
            except Exception as e:
                msg = (
                    f"the schedule contains a "
                    f"Field that is not understood: '{field}'"
                )
                raise ValueError(msg, e)
            else:
                return date
        else:
            return date

    @staticmethod
    def _parse_fancy_string(field, start_date):
        """Parse cases such as `3rd Monday in February` or `Last Weekday In Month`.

        Args:
            start_date:
            field (str): The EnergyPlus Field Contents

        Returns:
            (datetime): The datetime object
        """
        import re

        # split the string at the term ' in '
        time, month = field.lower().split(" in ")
        month = datetime.strptime(month, "%B").month

        # split the first part into nth and dayofweek
        nth, dayofweek = time.split(" ")
        if "last" in nth:
            nth = -1  # Use the last one
        else:
            nth = re.findall(r"\d+", nth)  # use the nth one
            nth = int(nth[0]) - 1  # python is zero-based

        weekday = {
            "monday": 0,
            "tuesday": 1,
            "wednesday": 2,
            "thursday": 3,
            "friday": 4,
            "saturday": 5,
            "sunday": 6,
        }

        # parse the dayofweek eg. monday
        dayofweek = weekday.get(dayofweek, 6)

        # create list of possible days using Calendar
        import calendar

        c = calendar.Calendar(firstweekday=start_date.weekday())
        monthcal = c.monthdatescalendar(start_date.year, month)

        # iterate though the month and get the nth weekday
        date = [
            day
            for week in monthcal
            for day in week
            if day.weekday() == dayofweek and day.month == month
        ][nth]
        return datetime(date.year, date.month, date.day)

    @staticmethod
    def special_day(schedule_epbunch, field, slicer_, strict, start_date):
        """Try to get the RunPeriodControl:SpecialDays for the corresponding DayType.

        Args:
            start_date:
            strict:
            field:
            slicer_:
        """
        sp_slicer_ = slicer_.copy()
        sp_slicer_.loc[:] = False
        special_day_types = ["holiday", "customday1", "customday2"]

        dds = schedule_epbunch.theidf.idfobjects["RunPeriodControl:SpecialDays".upper()]
        dd = [
            dd
            for dd in dds
            if dd.Special_Day_Type.lower() == field
            or dd.Special_Day_Type.lower() in special_day_types
        ]
        if len(dd) > 0:
            for dd in dd:
                # can have more than one special day types
                field = dd.Start_Date
                special_day_start_date = _ScheduleParser._date_field_interpretation(
                    field, start_date
                )
                duration = int(dd.Duration)
                to_date = (
                    special_day_start_date
                    + timedelta(days=duration)
                    + timedelta(hours=-1)
                )

                sp_slicer_.loc[special_day_start_date:to_date] = True
            return sp_slicer_
        elif not strict:
            return sp_slicer_
        else:
            msg = (
                'Could not find a "SizingPeriod:DesignDay" object '
                'needed for schedule with Day Type "{}"'.format(field.capitalize())
            )
            raise ValueError(msg)

    @staticmethod
    def design_day(schedule_epbunch, field, slicer_, start_date, strict):
        """Try to get the SizingPeriod:DesignDay for the corresponding Day Type.

        Args:
            strict:
            start_date:
            schedule_epbunch:
            field:
            slicer_:
        """
        sp_slicer_ = slicer_.copy()
        sp_slicer_.loc[:] = False
        dds = schedule_epbunch.theidf.idfobjects["SizingPeriod:DesignDay".upper()]
        dd = [dd for dd in dds if dd.Day_Type.lower() == field]
        if len(dd) > 0:
            for dd in dd:
                # should have found only one design day matching the Day Type
                month = dd.Month
                day = dd.Day_of_Month
                data = str(month) + "/" + str(day)
                ep_start_date = _ScheduleParser._date_field_interpretation(
                    data, start_date
                )
                ep_orig = datetime(start_date.year, 1, 1)
                days_to_speciald = (ep_start_date - ep_orig).days
                duration = 1  # Duration of 1 day
                from_date = start_date + timedelta(days=days_to_speciald)
                to_date = from_date + timedelta(days=duration) + timedelta(hours=-1)

                sp_slicer_.loc[from_date:to_date] = True
            return sp_slicer_
        elif not strict:
            return sp_slicer_
        else:
            msg = (
                f"Could not find a 'SizingPeriod:DesignDay' object "
                f"needed for schedule with Day Type '{field.capitalize()}'"
            )
            raise ValueError(msg)
            data = [dd[0].Month, dd[0].Day_of_Month]
            date = "/".join([str(item).zfill(2) for item in data])
            date = _ScheduleParser._date_field_interpretation(date, start_date)
            return lambda x: x.index == date


class Schedule:
    """Class handling any EnergyPlus schedule object."""

    def __init__(
        self,
        Name: str,
        start_day_of_the_week: FrozenSet[Literal[0, 1, 2, 3, 4, 5, 6]] = 0,
        strict: bool = False,
        Type: Union[str, ScheduleTypeLimits] = None,
        Values: Union[List[Union[int, float]], np.ndarray] = None,
        **kwargs,
    ):
        """Initialize object.

        Args:
            Name (str): The schedule name in the idf model.
            start_day_of_the_week (int): 0-based day of week (Monday=0). Default is
                None which looks for the start day in the IDF model.
            strict (bool): if True, schedules that have the Field-Sets such as
                Holidays and CustomDay will raise an error if they are absent
                from the IDF file. If False, any missing qualifiers will be
                ignored.
            Type (str, ScheduleTypeLimits): This field contains a reference to the
                Schedule Type Limits object. If found in a list of Schedule Type
                Limits (see above), then the restrictions from the referenced
                object will be used to validate the current field values. If None,
                no validation will occur.
            Values (ndarray): A 24 or 8760 list of schedule values.
            **kwargs:
        """
        try:
            super(Schedule, self).__init__(Name, **kwargs)
        except Exception:
            pass  # todo: make this more robust
        self.Name = Name
        self.strict = strict
        self.startDayOfTheWeek = start_day_of_the_week
        self.year = get_year_for_first_weekday(self.startDayOfTheWeek)
        self.Values = Values
        self.Type = Type

    @property
    def Type(self):
        """Get or set the schedule type limits object. Can be None."""
        return self._schedule_type_limits

    @Type.setter
    def Type(self, value):
        if isinstance(value, str):
            if "fraction" in value.lower():
                value = ScheduleTypeLimits("Fraction", 0, 1)
            elif value.lower() == "temperature":
                value = ScheduleTypeLimits("Temperature", -100, 100)
            elif value.lower() == "on/off":
                value = ScheduleTypeLimits("On/Off", 0, 1, "Discrete", "availability")
            else:
                value = ScheduleTypeLimits("Fraction", 0, 1)
                log(
                    f"'{value}' is not a known ScheduleTypeLimits for "
                    f"{self.Name}. Please instantiate the object before "
                    f"passing as the 'Type' parameter.",
                    level=lg.WARNING,
                )
            assert isinstance(value, ScheduleTypeLimits), value
        self._schedule_type_limits = value

    @property
    def Values(self):
        """Get or set the list of schedule values."""
        return self._values

    @Values.setter
    def Values(self, value):
        if isinstance(value, np.ndarray):
            assert value.ndim == 1, value.ndim
            value = value.tolist()
        self._values = validators.iterable(value, allow_empty=True)

    @property
    def Name(self):
        """Get or set the name of the schedule."""
        return self._name

    @Name.setter
    def Name(self, value):
        self._name = value

    @classmethod
    def from_values(
        cls,
        Name: str,
        Values: List[Union[float, int]],
        Type: str = "Fraction",
        **kwargs,
    ):
        """Create a Schedule from a list of Values."""
        return cls(Name=Name, Values=Values, Type=Type, **kwargs)

    @classmethod
    def from_epbunch(cls, epbunch, strict=False, Type=None, **kwargs):
        """Create a Schedule from an epbunch.

        Args:
            epbunch:
            strict:
            **kwargs:

        """
        if Type is None:
            try:
                type_limit_ep = epbunch.get_referenced_object(
                    "Schedule_Type_Limits_Name"
                )
                Type = ScheduleTypeLimits.from_epbunch(type_limit_ep)
            except (BadEPFieldError, AttributeError):
                pass
        name = epbunch.Name
        start_day_of_the_week = epbunch.theidf.day_of_week_for_start_day
        start_date = datetime(get_year_for_first_weekday(start_day_of_the_week), 1, 1)

        schedule = cls(
            Name=kwargs.pop("Name", name),
            epbunch=epbunch,
            start_day_of_the_week=kwargs.pop(
                "start_day_of_the_week", start_day_of_the_week
            ),
            Type=Type,
            DataSource=kwargs.pop("DataSource", epbunch.theidf.name),
            Values=np.array(
                _ScheduleParser.get_schedule_values(
                    epbunch, start_date=start_date, strict=strict
                )
            ),
            **kwargs,
        )
        return schedule

    @classmethod
    def constant_schedule(cls, value=1, Name="AlwaysOn", Type="Fraction", **kwargs):
        """Initialize a schedule with a constant value for the whole year.

        Defaults to a schedule with a value of 1, named 'AlwaysOn'.

        Args:
            value (float, optional): The value for the constant schedule.
                Defaults to 1.
            Name (str, optional): The name of the schedule. Defaults to Always
                On.
            **kwargs:
        """
        return cls.from_values(
            Name=Name,
            Values=np.ones((8760,)) * value,
            Type=Type,
            **kwargs,
        )

    @property
    def all_values(self) -> np.ndarray:
        """Return numpy array of schedule Values."""
        return np.array(self._values)

    @all_values.setter
    def all_values(self, value):
        self._values = validators.iterable(value, maximum_length=8760)

    @property
    def max(self):
        """Get the maximum value of the schedule."""
        return max(self.all_values)

    @property
    def min(self):
        """Get the minimum value of the schedule."""
        return min(self.all_values)

    @property
    def mean(self):
        """Get the mean value of the schedule."""
        return np.average(self.all_values)

    @property
    def series(self):
        """Return an :class:`EnergySeries`."""
        index = pd.date_range(
            start=self.startDate, periods=self.all_values.size, freq="1H"
        )
        if self.Type is not None:
            units = self.Type.UnitType
        else:
            units = None
        return EnergySeries(self.all_values, index=index, name=self.Name, units=units)

    @staticmethod
    def get_schedule_type_limits_name(epbunch):
        """Return the Schedule Type Limits name associated to this schedule."""
        try:
            schedule_limit_name = epbunch.Schedule_Type_Limits_Name
        except Exception:
            return "unknown"
        else:
            return schedule_limit_name

    @property
    def startDate(self):
        """Get the start date of the schedule. Satisfies `startDayOfTheWeek`."""
        year = get_year_for_first_weekday(self.startDayOfTheWeek)
        return datetime(year, 1, 1)

    def scale(self, diversity=0.1):
        """Scale the schedule values by a diversity factor around the average."""
        average = np.average(self.Values)
        new_values = ((average - self.Values) * diversity) + self.Values

        self.Values = new_values
        return self

    def replace(self, new_values: Union[pd.Series]):
        """Replace values with new values while keeping the full load hours constant.

        Time steps that are not specified in `new_values` will be adjusted to keep
        the full load hours of the schedule constant. No check whether the new schedule
        stays between the bounds set by self.Type is done. Be aware.

        """
        assert isinstance(new_values.index, pd.DatetimeIndex), (
            "The index of `new_values` must be a `pandas.DatetimeIndex`. Instead, "
            f"`{type(new_values.index)}` was provided."
        )
        assert not self.series.index.difference(new_values.index).empty, (
            "There is no overlap between self.index and new_values.index. Please "
            "check your dates."
        )

        # create a copy of self.series as orig.
        orig = self.series.copy()

        new_data = new_values.values

        # get the new_values index
        idx = new_values.index

        # compute the difference in values with the original data and the new data.
        diff = orig.loc[idx] - new_data.reshape(-1)

        # replace the original data with new values at their location.
        orig.loc[idx] = new_values

        # adjust remaining time steps with the average difference. Inplace.
        orig.loc[orig.index.difference(idx)] += diff.sum() / len(
            orig.index.difference(idx)
        )
        new = orig

        # assert the sum has not changed as a sanity check.
        np.testing.assert_array_almost_equal(self.series.sum(), new.sum())

        # replace values of self with new values.
        self.Values = new.tolist()

    def plot(self, **kwargs):
        """Plot the schedule. Implements the .loc accessor on the series object.

        Notes:
            Plotting can also be acheived through the series property:
            `Schedule.series.plot()`.

        Examples:
            >>> from archetypal import IDF
            >>> idf = IDF()
            >>> epbunch = idf.schedules_dict["NECB-A-Thermostat Setpoint-Heating"]
            >>> s = Schedule.from_epbunch(epbunch)
            >>>     )
            >>> s.plot(drawstyle="steps-post")

        Args:
            **kwargs (dict): keyword arguments passed to :meth:`EnergySeries.plot`.
        """
        return self.series.plot(**kwargs)

    def plot2d(self, **kwargs):
        """Plot the carpet plot of the schedule."""
        if self.Type is not None:
            vmin = self.Type.LowerLimit
            vmax = self.Type.UpperLimit
        else:
            vmin, vmax = (None, None)

        pretty_plot_kwargs = {
            "cmap": "Greys",
            "show": True,
            "figsize": (7, 2),
            "dpi": 72,
            "vmin": vmin,
            "vmax": vmax,
        }
        pretty_plot_kwargs.update(kwargs)
        return self.series.plot2d(**pretty_plot_kwargs)

    plot2d.__doc__ += EnergySeries.plot2d.__doc__

    def to_year_week_day(self):
        """Convert to three-tuple epbunch given an idf model.

        Returns 'Schedule:Year', 'Schedule:Week:Daily' and 'Schedule:Day:Hourly'
        representations.

        Returns:
            3-element tuple containing

            - **yearly** (*Schedule*): The yearly schedule object
            - **weekly** (*list of Schedule*): The list of weekly schedule
              objects
            - **daily** (*list of Schedule*):The list of daily schedule objects
        """
        from archetypal.template.schedule import (
            DaySchedule,
            WeekSchedule,
            YearSchedule,
            YearSchedulePart,
        )

        full_year = np.array(self.all_values)  # array of shape (8760,)

        # reshape to (365, 24)
        Values = full_year.reshape(-1, 24)  # shape (365, 24)

        # create unique days
        unique_days, nds = np.unique(Values, axis=0, return_inverse=True)

        ep_days = []
        dict_day = {}
        for count_day, unique_day in enumerate(unique_days):
            name = f"d_{self.Name}_{count_day:02d}"
            dict_day[name] = unique_day

            # Create idf_objects for schedule:day:hourly
            ep_day = DaySchedule(
                Name=name, Type=self.Type, Values=[unique_day[i] for i in range(24)]
            )
            ep_days.append(ep_day)

        # create unique weeks from unique days
        try:
            unique_weeks, nwsi, nws, count = np.unique(
                full_year[: 364 * 24, ...].reshape(-1, 168),
                return_index=True,
                axis=0,
                return_inverse=True,
                return_counts=True,
            )
        except ValueError:
            raise ValueError(
                "Looks like the idf model needs to be rerun with 'annual=True'"
            )

        # We use the calendar module to set the week days order
        import calendar

        # initialize the calendar object
        c = calendar.Calendar(firstweekday=self.startDayOfTheWeek)

        # Appending unique weeks in dictionary with name and values of weeks as
        # keys
        # {'name_week': {'dayName':[]}}
        dict_week = {}
        for count_week, unique_week in enumerate(unique_weeks):
            week_id = f"w_{self.Name}_{count_week:02d}"
            dict_week[week_id] = {}
            for i, day in zip(list(range(0, 7)), list(c.iterweekdays())):
                day_of_week = unique_week[..., i * 24 : (i + 1) * 24]
                for key, ep_day in zip(dict_day, ep_days):
                    if (day_of_week == dict_day[key]).all():
                        dict_week[week_id]["day_{}".format(day)] = ep_day

        # Create idf_objects for schedule:week:daily

        # Create ep_weeks list and iterate over dict_week
        ep_weeks = {}
        for week_id in dict_week:
            ep_week = WeekSchedule(
                Name=week_id,
                Days=[
                    dict_week[week_id]["day_{}".format(day_num)]
                    for day_num in c.iterweekdays()
                ],
            )
            ep_weeks[week_id] = ep_week

        blocks = {}
        from_date = datetime(self.year, 1, 1)
        bincount = [sum(1 for _ in group) for key, group in groupby(nws + 1) if key]
        week_order = {
            i: v
            for i, v in enumerate(
                np.array([key for key, group in groupby(nws + 1) if key]) - 1
            )
        }
        for i, (week_n, count) in enumerate(zip(week_order, bincount)):
            week_id = list(dict_week)[week_order[i]]
            to_date = from_date + timedelta(days=int(count * 7), hours=-1)
            blocks[i] = YearSchedulePart(
                FromDay=from_date.day,
                FromMonth=from_date.month,
                ToDay=to_date.day,
                ToMonth=to_date.month,
                Schedule=ep_weeks[week_id],
            )
            from_date = to_date + timedelta(hours=1)

            # If this is the last block, force end of year
            if i == len(bincount) - 1:
                blocks[i].ToDay = 31
                blocks[i].ToMonth = 12

        ep_year = YearSchedule(self.Name, Type=self.Type, Parts=list(blocks.values()))

        return ep_year, list(ep_weeks.values()), ep_days

    def __len__(self):
        """Get the length of all values of the schedule."""
        return len(self.all_values)

    def __add__(self, other):
        """Add self and other."""
        if isinstance(other, Schedule):
            return self.all_values + other.all_values
        elif isinstance(other, list):
            return self.all_values + other
        else:
            raise NotImplementedError

    def __sub__(self, other):
        """Subtract self and other."""
        if isinstance(other, Schedule):
            return self.all_values - other.all_values
        elif isinstance(other, list):
            return self.all_values - other
        else:
            raise NotImplementedError

    def __mul__(self, other):
        """Multiply self with other."""
        if isinstance(other, Schedule):
            return self.all_values * other.all_values
        elif isinstance(other, list):
            return self.all_values * other
        else:
            raise NotImplementedError

    def _repr_svg_(self):
        """SVG representation for iPython notebook."""
        if self.Type is not None:
            vmin = self.Type.LowerLimit
            vmax = self.Type.UpperLimit
        else:
            vmin, vmax = (None, None)
        fig, ax = self.series.plot2d(
            cmap="Greys", show=False, figsize=(7, 2), dpi=72, vmin=vmin, vmax=vmax
        )
        f = io.BytesIO()
        fig.savefig(f, format="svg")
        return f.getvalue()

    def combine(self, other, weights=None, quantity=None):
        """Combine two schedule objects together.

        Args:
            other (Schedule): the other Schedule object to combine with.
            weights (list-like, optional): A list-like object of len 2. If None,
                equal weights are used.
            quantity: scalar value that will be multiplied by self before the
                averaging occurs. This ensures that the resulting schedule
                returns the correct integrated value.

        Returns:
            (Schedule): the combined Schedule object.
        """
        # Check if other is None. Simply return self
        if not other:
            return self

        if not self:
            return other

        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # check if the schedule is the same

        if all(self.all_values == other.all_values):
            return self
        if not weights:
            weights = [1, 1]
        new_values = np.average(
            [self.all_values, other.all_values], axis=0, weights=weights
        )

        # the new object's name
        name = "+".join([self.Name, other.Name])

        new_obj = self.__class__(name, value=new_values)

        return new_obj


def _conjunction(*conditions, logical=np.logical_and):
    return functools.reduce(logical, conditions)


def _separator(sep):
    if sep == "Comma":
        return ","
    elif sep == "Tab":
        return "\t"
    elif sep == "Fixed":
        return None
    elif sep == "Semicolon":
        return ";"
    else:
        return ","


def _how(how):
    if how.lower() == "average":
        return "mean"
    elif how.lower() == "linear":
        return "interpolate"
    elif how.lower() == "no":
        return "max"
    else:
        return "max"


def get_year_for_first_weekday(weekday: FrozenSet[Literal[0, 1, 2, 3, 4, 5, 6]] = 0):
    """Get the year that starts on 'weekday', eg. Monday=0.

    Args:
        weekday (int): 0-based day of week (Monday=0). Default is
            None which looks for the start day in the IDF model.

    Returns:
        (int): The year number for which the first starts on :attr:`weekday`.
    """
    import calendar

    if weekday > 6:
        raise ValueError("weekday must be between 0 and 6")
    year = 2020
    not_found = True
    while not_found:
        firstday = calendar.weekday(year, 1, 1)
        if firstday == weekday and not calendar.isleap(year):
            not_found = False
        else:
            year = year - 1
    return year



################################################################################
# Module: cli.py
# Description: Implements archetypal functions as Command Line Interfaces
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################
import os
import time

import click
from path import Path

from archetypal import __version__, settings
from archetypal.idfclass import IDF
from archetypal.umi_template import UmiTemplateLibrary
from archetypal.utils import config, docstring_parameter, log, parallel_process, timeit

from .eplus_interface.exceptions import EnergyPlusVersionError
from .eplus_interface.version import EnergyPlusVersion

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


class CliConfig(object):
    def __init__(self):
        self.data_folder = settings.data_folder
        self.logs_folder = settings.logs_folder
        self.imgs_folder = settings.imgs_folder
        self.cache_folder = settings.cache_folder
        self.cache_responses = settings.cache_responses
        self.log_file = settings.log_file
        self.log_console = settings.log_console
        self.log_level = settings.log_level
        self.log_name = settings.log_name
        self.log_filename = settings.log_filename
        self.useful_idf_objects = settings.useful_idf_objects
        self.ep_version = settings.ep_version


pass_config = click.make_pass_decorator(CliConfig, ensure=True)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--data-folder",
    type=click.Path(),
    help="where to save and load data files",
    default=settings.data_folder,
)
@click.option(
    "--logs-folder",
    type=click.Path(),
    help="where to write the log files",
    default=settings.logs_folder,
)
@click.option(
    "--imgs-folder",
    type=click.Path(),
    help="where to save figures",
    default=settings.imgs_folder,
)
@click.option(
    "--cache-folder",
    type=click.Path(),
    help="where to save the simluation results",
    default=settings.cache_folder,
)
@click.option(
    "-c",
    "--cache-responses",
    is_flag=True,
    default=False,
    help="Use a local cache to save/retrieve DataPortal API calls for the same "
    "requests.",
)
@click.option(
    "-l",
    "--log-file",
    "log_file",
    is_flag=True,
    help="save log output to a log file in logs_folder",
    default=settings.log_file,
)
@click.option(
    "--silent",
    "-s",
    "log_console",
    is_flag=True,
    default=True,
    help="print log output to the console",
)
@click.option(
    "--log-level",
    type=click.INT,
    help="CRITICAL=50, ERROR=40, WARNING=30, INFO=20, DEBUG=10",
    default=settings.log_level,
)
@click.option("--log-name", help="name of the logger", default=settings.log_name)
@click.option(
    "--log-filename", help="name of the log file", default=settings.log_filename
)
@click.option(
    "--ep_version",
    type=click.STRING,
    default=settings.ep_version,
    help='the EnergyPlus version to use. eg. "{}"'.format(settings.ep_version),
)
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    default=False,
    help="Will break on any exception. Useful when debugging",
)
@pass_config
def cli(
    cli_config,
    data_folder,
    logs_folder,
    imgs_folder,
    cache_folder,
    cache_responses,
    log_file,
    log_console,
    log_level,
    log_name,
    log_filename,
    ep_version,
    debug,
):
    """archetypal: Retrieve, construct, simulate, convert and analyse building
    simulation templates

    Visit archetypal.readthedocs.io for the online documentation.

    """
    cli_config.data_folder = data_folder
    cli_config.logs_folder = logs_folder
    cli_config.imgs_folder = imgs_folder
    cli_config.cache_folder = cache_folder
    cli_config.cache_responses = cache_responses
    cli_config.log_file = log_file
    cli_config.log_console = log_console
    cli_config.log_level = log_level
    cli_config.log_name = log_name
    cli_config.log_filename = log_filename
    cli_config.ep_version = ep_version
    cli_config.debug = debug
    # apply new config params
    config(**cli_config.__dict__)


@timeit
@cli.command()
@click.argument("idf", nargs=-1, required=True)
@click.option(
    "-o",
    "--output",
    type=click.Path(dir_okay=True, writable=True),
    default="myumitemplate.json",
)
@click.option(
    "--weather",
    "-w",
    help="EPW weather file path",
    default=EnergyPlusVersion.current().current_install_location
    / "WeatherData"
    / "USA_IL_Chicago-OHare.Intl.AP.725300_TMY3.epw",
    show_default=True,
)
@click.option(
    "-p",
    "--parallel",
    "cores",
    default=-1,
    help="Specify number of cores to run in parallel",
)
@click.option(
    "-z",
    "--all_zones",
    is_flag=True,
    default=False,
    help="Include all zones in the " "output template",
)
@click.option(
    "-v",
    "--version",
    "as_version",
    default=settings.ep_version,
    help="EnergyPlus version to upgrade to - e.g., '9-2-0'",
)
@click.pass_context
def reduce(ctx, idf, output, weather, cores, all_zones, as_version):
    """Convert EnergyPlus models to an Umi Template Library by using the model
    complexity reduction algorithm.

    IDF can be a file path or a directory. In case of a directory, all *.idf
    files will be matched in the directory and subdirectories (recursively). Mix &
    match is ok (see example below).
    OUTPUT is the output file
    name (or path) to write to. Optional.

    Example: % archetypal -csl reduce "." "elsewhere/model1.idf" -w "weather.epw"

    """
    output = Path(output)
    name = output.stem
    ext = output.ext if output.ext == ".json" else ".json"
    dir_ = output.dirname()

    file_paths = list(set_filepaths(idf))
    file_list = "\n".join(
        [f"{i}. " + str(file.name) for i, file in enumerate(file_paths)]
    )
    log(
        f"executing {len(file_paths)} file(s):\n{file_list}",
    )
    weather, *_ = set_filepaths([weather])
    log(f"using the '{weather.basename()}' weather file\n")

    # Call UmiTemplateLibrary constructor with list of IDFs
    template = UmiTemplateLibrary.from_idf_files(
        file_paths,
        weather=weather,
        name=name,
        processors=cores,
        keep_all_zones=all_zones,
        as_version=as_version,
        annual=True,
    )
    # Save json file
    final_path: Path = dir_ / name + ext
    template.save(path_or_buf=final_path)
    log(
        f"Successfully created template file at {final_path.abspath()}",
    )


def validate_energyplusversion(ctx, param, value):
    try:
        return EnergyPlusVersion(value)
    except EnergyPlusVersionError:
        raise click.BadParameter("invalid energyplus version")


def validate_paths(ctx, param, value):
    try:
        file_paths = set_filepaths(value)
        file_list = "\n".join(
            [f"{i}. " + str(file.name) for i, file in enumerate(file_paths)]
        )
        return file_paths, file_list
    except FileNotFoundError:
        raise click.BadParameter("no files were found.")


@cli.command()
@click.argument("idf", nargs=-1, required=True, callback=validate_paths)
@click.option(
    "-v",
    "--version",
    "to_version",
    default=settings.ep_version,
    help="EnergyPlus version to upgrade to - e.g., '9-2-0'",
    callback=validate_energyplusversion,
)
@click.option(
    "-p",
    "--parallel",
    "cores",
    default=-1,
    help="Specify number of cores to run in parallel",
)
@click.option(
    "-y",
    "yes",
    is_flag=True,
    help="Suppress confirmation prompt, when overwriting files.",
)
@docstring_parameter(arversion=__version__, ep_version=settings.ep_version)
def transition(idf, to_version, cores, yes):
    """Upgrade an IDF file to a newer version.

    IDF can be a file path or a directory. In case of a directory, all *.idf
    files will be found in the directory and subdirectories (recursively). Mix &
    match is ok (see example below).

    Example: % archetypal -csl transition "." "elsewhere/model1.idf"

    archetypal will look in the current working directory (".") and find any
    *.idf files and also run the model located at "elsewhere/model1.idf".

    Note: The latest version archetypal v{arversion} can upgrade to is
    {ep_version}.

    """
    file_paths, file_list = idf
    log(
        f"executing {len(file_paths)} file(s):\n{file_list}",
    )
    if not yes:
        overwrite = click.confirm("Would you like to overwrite the file(s)?")
    else:
        overwrite = False
    start_time = time.time()

    to_version = to_version.dash
    rundict = {
        file: dict(
            idfname=file,
            as_version=to_version,
            check_required=False,
            check_length=False,
            overwrite=overwrite,
            prep_outputs=False,
        )
        for i, file in enumerate(file_paths)
    }
    results = parallel_process(
        rundict,
        IDF,
        processors=cores,
        show_progress=True,
        position=0,
        debug=False,
    )

    # Save results to file (overwriting if True)
    file_list = []
    for idf in results.values():
        if isinstance(idf, IDF):
            if overwrite:
                file_list.append(idf.original_idfname)
                idf.saveas(str(idf.original_idfname))
            else:
                full_path = (
                    idf.original_idfname.dirname() / idf.original_idfname.stem
                    + f"V{to_version}.idf"
                )
                file_list.append(full_path)
                idf.saveas(full_path)
    log(
        f"Successfully transitioned to version '{to_version}' in "
        f"{time.time() - start_time:,.2f} seconds for file(s):\n" + "\n".join(file_list)
    )


def set_filepaths(idf):
    """Simplifies file-like paths, dir-like paths and Paths with wildcards. A
    list of unique paths is returned. For directories, Path.walkfiles("*.idfs")
    returns IDF files. For Paths with wildcards, glob(Path) is used to return
    whatever the pattern defines.

    Args:
        idf (list of (str or Path) or tuple of (str or Path)): A list of path-like
            objects. Can contain wildcards.

    Returns:
        set of Path: The set of a list of paths
    """
    if not isinstance(idf, (list, tuple)):
        raise ValueError("A list must be passed")
    idf = tuple(Path(file_or_path).expand() for file_or_path in idf)  # make Paths
    file_paths = ()  # Placeholder for tuple of paths
    for file_or_path in idf:
        if file_or_path.isfile():  # if a file, concatenate into file_paths
            file_paths += tuple([file_or_path])
        elif file_or_path.isdir():  # if a directory, walkdir (recursive) and get *.idf
            file_paths += tuple(file_or_path.walkfiles("*.idf"))
        else:
            # has wildcard
            excluded_dirs = [
                settings.cache_folder,
                settings.data_folder,
                settings.imgs_folder,
                settings.logs_folder,
            ]
            top = file_or_path.abspath().dirname()
            for root, dirs, files in walkdirs(top, excluded_dirs):
                pattern = file_or_path.basename()
                file_paths += tuple(Path(root).files(pattern))

    file_paths = set([f.relpath().expand() for f in file_paths])  # Only keep unique
    # values
    if file_paths:
        return file_paths
    else:
        raise FileNotFoundError


def walkdirs(top, excluded):
    for root, dirs, files in os.walk(top, topdown=True):
        yield root, dirs, files
        dirs[:] = [d for d in dirs if (Path(root) / d) not in excluded]
